// HC11/CS8900 ethernet IP stack
//
// $Id: hc11.c,v 3.0 2002/04/10 11:59:45 tom Exp tom $
//
// Copyright (C) 2000-1 Tom Parker (tom-hc11@wiresncode.com)
// Serial code Copyright 1999, 2000 Free Software Foundation, Inc.
//         Written by Stephane Carrez (stcarrez@worldnet.fr)
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
//
// Author's note: This is free software, for commerical use I would
// appreciate some recognition though.  If you are using this and like
// it send me some email or a postcard.  Code changes for inclusion
// are also welcome.
//
// Tom Parker, PO Box 2165, Fitzroy, VIC, 3065, AUSTRALIA (tom-hc11@wiresncode.com)
//

#include "ether.h"
#include "timer.h"
#include "../m68hc11/ports_def.h"
#include "ports_def_f1.h"
#include "cs8900reg.h"
#include "serial.h"


#ifdef BUFFALO
#define PVBASE 0x7e00
#define PVRTI  0x27
#endif

#define CS8900_OFFSET	0x60	/* CS8900 mapped from 0x60 - 0x6f */

#define M6811_CPU_CLOCK (16000000L)
#define M6811_DEF_BAUD  (0x30)


// divider for per second interrupt
u_int16_t timerTicks;


void time_of_day() __attribute__((interrupt));

//-----------------------------------------------------------------
// HC11 initialisation
void
cpu_init()
{

#ifdef BUFFALO
   u_int8_t* memptr = (u_int8_t*) PVBASE + PVRTI;

   // fill buffalo psuedo vector table with a jump to the time_of_day handler
   *(memptr++) = 0x7e;
   *(memptr++) = (u_int16_t) time_of_day >> 8;
   *(memptr) = (u_int16_t) time_of_day & 0xff;
#endif

   // CS follows E with 1 cycle stretch
   _io_ports[M6811F1_CSSTRH] |=  M6811F1_IO1SB;
   _io_ports[M6811F1_CSSTRH] &= ~M6811F1_IO1SA;
   
   // enable CS1, disable CS2
   _io_ports[M6811F1_CSCTL] |= M6811F1_IO1EN;
   _io_ports[M6811F1_CSCTL] &= ~M6811F1_IO2EN;
   
   // A/D converter off
   _io_ports[M6811_OPTION] |= ~M6811_ADPU;

   // Port G outputs
   _io_ports[M6811F1_DDRG] &= 0x0f;

   // config
   _io_ports[M6811_PACTL] |= M6811_RTR0;
   _io_ports[M6811_PACTL] |= M6811_RTR1;
   _io_ports[M6811_TMSK2] |= M6811_RTII;

#ifndef BUFFALO
   // serial port
   _io_ports[M6811_BAUD] = M6811_DEF_BAUD;
   _io_ports[M6811_SCCR1] = 0;
   _io_ports[M6811_SCCR2] = 0xc;
#endif

}

//-----------------------------------------------------------------
// Get node ID from LSB of port E
u_int8_t
cpu_get_nodeid()
{
    return _io_ports[M6811_PORTE] & 0xf;
}

//-----------------------------------------------------------------
// Send a character out the serial port
void
cpu_serial_send(char c)
{
    while (!(_io_ports[M6811_SCSR] & M6811_TDRE));
    _io_ports[M6811_SCDR] = c;
    _io_ports[M6811_SCCR2] |= M6811_TE;
}

//-----------------------------------------------------------------
// read packet page register from cs8900
u_int16_t
cpu_cs8900_pp_reg_read(u_int16_t reg)
{
   _io_ports[CS8900_PP_ADDR_LSB + CS8900_OFFSET] = (u_int8_t) reg & 0xff;
   _io_ports[CS8900_PP_ADDR_MSB + CS8900_OFFSET] = (u_int8_t) (reg >> 8) & 0xff;

   return (_io_ports[CS8900_PPDATA_PORT0_LSB + CS8900_OFFSET] |
	   (_io_ports[CS8900_PPDATA_PORT0_MSB + CS8900_OFFSET] << 8));
}

//-----------------------------------------------------------------
// write to packet page register in cs8900
void
cpu_cs8900_pp_reg_write(u_int16_t reg, u_int16_t data)
{
   _io_ports[CS8900_PP_ADDR_LSB + CS8900_OFFSET] = (u_int8_t) reg & 0xff;
   _io_ports[CS8900_PP_ADDR_MSB + CS8900_OFFSET] = (u_int8_t) (reg >> 8) & 0xff;

   _io_ports[CS8900_PPDATA_PORT0_LSB + CS8900_OFFSET] = data & 0xff;
   _io_ports[CS8900_PPDATA_PORT0_MSB + CS8900_OFFSET] = (data >> 8) & 0xff;
}

//-----------------------------------------------------------------
// read from 8 bit cs8900 register
u_int8_t 
cpu_cs8900_reg_read(u_int8_t reg)
{
    return _io_ports[reg + CS8900_OFFSET];
}

//-----------------------------------------------------------------
// write to 8 bit cs8900 register
void 
cpu_cs8900_reg_write(const u_int8_t reg, const u_int8_t data)
{
    _io_ports[reg + CS8900_OFFSET] = data;
}

//-----------------------------------------------------------------
// evaluate 16 bit one's complement checksum over buffer
// aBuffer points to ip packet, aLength is length of ip payload
u_int16_t 
cpu_checksum(u_int16_t* aBuffer, u_int16_t aLength)
{
   register u_int8_t i;
   XDATA u_int32_t newcksum;

   serial_print_debug('+');

   // do we need to includes portions of the IP header ie for TCP/UDP?
   if ((aLength & CPU_CHECKSUM_INC_IP) != 0)
   {
      serial_print_debug('1');

      // remove flag
      aLength &= ~CPU_CHECKSUM_INC_IP;

      // start checksum with protocol ID
      newcksum = *((u_int8_t *)aBuffer + 9);

      // add source and dest IP addresses
      newcksum += *(aBuffer + 6);
      newcksum += *(aBuffer + 7);
      newcksum += *(aBuffer + 8);
      newcksum += *(aBuffer + 9);

      // add inner protocol aLength
      newcksum += (aLength - sizeof(struct ip));

      // skip ip header
      aBuffer += (sizeof(struct ip) / 2);
   }
   else
   {
      newcksum = 0;
   }

   if (aLength & 1)
   {
      aBuffer[aLength + 1] = 0;
      aLength++;
   }

   aLength = aLength >> 1;

   for (i = 0; i < aLength; i++, aBuffer++)
   {
      newcksum += ntohs(*aBuffer);
   }

   return htons(~((newcksum >> 16) + (newcksum & 0xffff)));
}

//-----------------------------------------------------------------
// Time of day ISR and maintenance
// 
// ISR will fire every 2.048 ms.  time_of_day is hooked to the RTI
// Real Time Interrupt in vectors.s for standalone or in hc11_init
// for buffalo rom
//
#define TIMER_INTS_PER_SECOND (1000000/16384)

void time_of_day(void)
{
   // fire every second
   if (++timerTicks == TIMER_INTS_PER_SECOND)
   {
      timerTicks = 0;
      serial_print_debug('!');
      timer_per_second();
   }

   // re-enable interrupt
   _io_ports[M6811_TFLG2] |= M6811_RTIF;
}
